/**
 * @license
 * Copyright 2014 Palantir Technologies, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
const utils = require("tsutils");
const ts = require("typescript");
const Lint = require("tslint");
const OPTION_CHECK_PARAMETERS = "check-parameters";
const OPTION_IGNORE_PATTERN = "ignore-pattern";
class Rule extends Lint.Rules.TypedRule {
    /* tslint:enable:object-literal-sort-keys */
    applyWithProgram(sourceFile, program) {
        const x = program.getCompilerOptions();
        if (x.noUnusedLocals && x.noUnusedParameters) {
            console.warn("WARNING: 'no-unused-variable' lint rule does not need to be set if " +
                "the 'no-unused-locals' and 'no-unused-parameters' compiler options are enabled.");
        }
        return this.applyWithFunction(sourceFile, (ctx) => walk(ctx, program, parseOptions(this.ruleArguments)));
    }
}
/* tslint:disable:object-literal-sort-keys */
Rule.metadata = {
    ruleName: "no-unused-variable",
    description: Lint.Utils.dedent `Disallows unused imports, variables, functions and
            private class members. Similar to tsc's --noUnusedParameters and --noUnusedLocals
            options, but does not interrupt code compilation.`,
    hasFix: true,
    optionsDescription: Lint.Utils.dedent `
            Three optional arguments may be optionally provided:

            * \`"check-parameters"\` disallows unused function and constructor parameters.
                * NOTE: this option is experimental and does not work with classes
                that use abstract method declarations, among other things.
            * \`{"ignore-pattern": "pattern"}\` where pattern is a case-sensitive regexp.
            Variable names that match the pattern will be ignored.`,
    options: {
        type: "array",
        items: {
            oneOf: [{
                    type: "string",
                    enum: ["check-parameters"],
                }, {
                    type: "object",
                    properties: {
                        "ignore-pattern": { type: "string" },
                    },
                    additionalProperties: false,
                }],
        },
        minLength: 0,
        maxLength: 3,
    },
    optionExamples: [true, [true, { "ignore-pattern": "^_" }]],
    type: "functionality",
    typescriptOnly: true,
    requiresTypeInfo: true,
};
exports.Rule = Rule;
function parseOptions(options) {
    const checkParameters = options.indexOf(OPTION_CHECK_PARAMETERS) !== -1;
    let ignorePattern;
    for (const o of options) {
        if (typeof o === "object") {
            // tslint:disable-next-line no-unsafe-any
            const ignore = o[OPTION_IGNORE_PATTERN];
            if (ignore != null) {
                ignorePattern = new RegExp(ignore);
                break;
            }
        }
    }
    return { checkParameters, ignorePattern };
}
function walk(ctx, program, { checkParameters, ignorePattern }) {
    const { sourceFile } = ctx;
    const unusedCheckedProgram = getUnusedCheckedProgram(program, checkParameters);
    const diagnostics = ts.getPreEmitDiagnostics(unusedCheckedProgram, sourceFile);
    const checker = unusedCheckedProgram.getTypeChecker(); // Doesn't matter which program is used for this.
    // If all specifiers in an import are unused, we elide the entire import.
    const importSpecifierFailures = new Map();
    for (const diag of diagnostics) {
        const kind = getUnusedDiagnostic(diag);
        if (kind === undefined) {
            continue;
        }
        const failure = ts.flattenDiagnosticMessageText(diag.messageText, "\n");
        if (kind === 0 /* VARIABLE_OR_PARAMETER */) {
            const importName = findImport(diag.start, sourceFile);
            if (importName !== undefined) {
                if (isImportUsed(importName, sourceFile, checker)) {
                    continue;
                }
                if (importSpecifierFailures.has(importName)) {
                    throw new Error("Should not get 2 errors for the same import.");
                }
                importSpecifierFailures.set(importName, failure);
                continue;
            }
        }
        if (ignorePattern) {
            const varName = /'(.*)'/.exec(failure)[1];
            if (ignorePattern.test(varName)) {
                continue;
            }
        }
        ctx.addFailureAt(diag.start, diag.length, failure);
    }
    if (importSpecifierFailures.size) {
        addImportSpecifierFailures(ctx, importSpecifierFailures, sourceFile);
    }
}
/**
 * Handle import-specifier failures separately.
 * - If all of the import specifiers in an import are unused, add a combined failure for them all.
 * - Unused imports are fixable.
 */
function addImportSpecifierFailures(ctx, failures, sourceFile) {
    forEachImport(sourceFile, (importNode) => {
        if (importNode.kind === ts.SyntaxKind.ImportEqualsDeclaration) {
            tryRemoveAll(importNode.name);
            return;
        }
        if (!importNode.importClause) {
            // Error node
            return;
        }
        const { name: defaultName, namedBindings } = importNode.importClause;
        if (namedBindings && namedBindings.kind === ts.SyntaxKind.NamespaceImport) {
            tryRemoveAll(namedBindings.name);
            return;
        }
        const allNamedBindingsAreFailures = !namedBindings || namedBindings.elements.every((e) => failures.has(e.name));
        if (namedBindings && allNamedBindingsAreFailures) {
            for (const e of namedBindings.elements) {
                failures.delete(e.name);
            }
        }
        if ((!defaultName || failures.has(defaultName)) && allNamedBindingsAreFailures) {
            if (defaultName) {
                failures.delete(defaultName);
            }
            removeAll(importNode, "All imports are unused.");
            return;
        }
        if (defaultName) {
            const failure = tryDelete(defaultName);
            if (failure !== undefined) {
                const start = defaultName.getStart();
                const end = namedBindings ? namedBindings.getStart() : importNode.moduleSpecifier.getStart();
                const fix = Lint.Replacement.deleteFromTo(start, end);
                ctx.addFailureAtNode(defaultName, failure, fix);
            }
        }
        if (namedBindings) {
            if (allNamedBindingsAreFailures) {
                const start = defaultName ? defaultName.getEnd() : namedBindings.getStart();
                const fix = Lint.Replacement.deleteFromTo(start, namedBindings.getEnd());
                const failure = "All named bindings are unused.";
                ctx.addFailureAtNode(namedBindings, failure, fix);
            }
            else {
                const { elements } = namedBindings;
                for (let i = 0; i < elements.length; i++) {
                    const element = elements[i];
                    const failure = tryDelete(element.name);
                    if (failure === undefined) {
                        continue;
                    }
                    const prevElement = elements[i - 1];
                    const nextElement = elements[i + 1];
                    const start = prevElement ? prevElement.getEnd() : element.getStart();
                    const end = nextElement && !prevElement ? nextElement.getStart() : element.getEnd();
                    const fix = Lint.Replacement.deleteFromTo(start, end);
                    ctx.addFailureAtNode(element.name, failure, fix);
                }
            }
        }
        function tryRemoveAll(name) {
            const failure = tryDelete(name);
            if (failure !== undefined) {
                removeAll(name, failure);
            }
        }
        function removeAll(errorNode, failure) {
            const fix = Lint.Replacement.deleteFromTo(importNode.getStart(), importNode.getEnd());
            ctx.addFailureAtNode(errorNode, failure, fix);
        }
    });
    if (failures.size) {
        throw new Error("Should have revisited all import specifier failures.");
    }
    function tryDelete(name) {
        const failure = failures.get(name);
        if (failure !== undefined) {
            failures.delete(name);
            return failure;
        }
        return undefined;
    }
}
/**
 * Ignore this import if it's used as an implicit type somewhere.
 * Workround for https://github.com/Microsoft/TypeScript/issues/9944
 */
function isImportUsed(importSpecifier, sourceFile, checker) {
    let symbol = checker.getSymbolAtLocation(importSpecifier);
    if (!symbol) {
        return false;
    }
    symbol = checker.getAliasedSymbol(symbol);
    if (!Lint.isSymbolFlagSet(symbol, ts.SymbolFlags.Type)) {
        return false;
    }
    return ts.forEachChild(sourceFile, function cb(child) {
        if (isImportLike(child)) {
            return false;
        }
        const type = getImplicitType(child, checker);
        // TODO: checker.typeEquals https://github.com/Microsoft/TypeScript/issues/13502
        if (type && checker.typeToString(type) === checker.symbolToString(symbol)) {
            return true;
        }
        return ts.forEachChild(child, cb);
    });
}
function getImplicitType(node, checker) {
    if ((utils.isPropertyDeclaration(node) || utils.isVariableDeclaration(node)) && !node.type) {
        return checker.getTypeAtLocation(node);
    }
    else if (utils.isSignatureDeclaration(node) && !node.type) {
        return checker.getSignatureFromDeclaration(node).getReturnType();
    }
    else {
        return undefined;
    }
}
function isImportLike(node) {
    return node.kind === ts.SyntaxKind.ImportDeclaration || node.kind === ts.SyntaxKind.ImportEqualsDeclaration;
}
function forEachImport(sourceFile, f) {
    return ts.forEachChild(sourceFile, (child) => {
        if (isImportLike(child)) {
            const res = f(child);
            if (res !== undefined) {
                return res;
            }
        }
        return undefined;
    });
}
function findImport(pos, sourceFile) {
    return forEachImport(sourceFile, (i) => {
        if (i.kind === ts.SyntaxKind.ImportEqualsDeclaration) {
            if (i.name.getStart() === pos) {
                return i.name;
            }
        }
        else {
            if (!i.importClause) {
                // Error node
                return undefined;
            }
            const { name: defaultName, namedBindings } = i.importClause;
            if (namedBindings && namedBindings.kind === ts.SyntaxKind.NamespaceImport) {
                const { name } = namedBindings;
                if (name.getStart() === pos) {
                    return name;
                }
                return undefined;
            }
            if (defaultName && defaultName.getStart() === pos) {
                return defaultName;
            }
            else if (namedBindings) {
                for (const { name } of namedBindings.elements) {
                    if (name.getStart() === pos) {
                        return name;
                    }
                }
            }
        }
        return undefined;
    });
}
function getUnusedDiagnostic(diag) {
    switch (diag.code) {
        case 6133:
            return 0 /* VARIABLE_OR_PARAMETER */; // "'{0}' is declared but never used.
        case 6138:
            return 1 /* PROPERTY */; // "Property '{0}' is declared but never used."
        default:
            return undefined;
    }
}
const programToUnusedCheckedProgram = new WeakMap();
function getUnusedCheckedProgram(program, checkParameters) {
    // Assuming checkParameters will always have the same value, so only lookup by program.
    let checkedProgram = programToUnusedCheckedProgram.get(program);
    if (checkedProgram) {
        return checkedProgram;
    }
    checkedProgram = makeUnusedCheckedProgram(program, checkParameters);
    programToUnusedCheckedProgram.set(program, checkedProgram);
    return checkedProgram;
}
function makeUnusedCheckedProgram(program, checkParameters) {
    const options = __assign({}, program.getCompilerOptions(), { noUnusedLocals: true }, (checkParameters ? { noUnusedParameters: true } : null));
    const sourceFilesByName = new Map(program.getSourceFiles().map((s) => [s.fileName, s]));
    // tslint:disable object-literal-sort-keys
    return ts.createProgram(Array.from(sourceFilesByName.keys()), options, {
        fileExists: (f) => sourceFilesByName.has(f),
        readFile(f) {
            const s = sourceFilesByName.get(f);
            return s.text;
        },
        getSourceFile: (f) => sourceFilesByName.get(f),
        getDefaultLibFileName: () => ts.getDefaultLibFileName(options),
        writeFile: () => { },
        getCurrentDirectory: () => "",
        getDirectories: () => [],
        getCanonicalFileName: (f) => f,
        useCaseSensitiveFileNames: () => true,
        getNewLine: () => "\n",
    });
    // tslint:enable object-literal-sort-keys
}
//# sourceMappingURL=CustomNoUnusedVariableRule.js.map